﻿-- function schedluer frame
local BMTSchedulerFrame = CreateFrame("Frame");
-- time since last call of the OnUpdate
local BMTElapsed = 0;
BMTScheduledTasks = {};
-- OnUpdate calls throttle
local BMT_ONUPDATE_SEC = 0.5;
-- set to true for additional dev messages
local BMTDebug = false;
--[===[@debug@
BMTDebug = true;
--@end-debug@]===]
-- this revision number
--BMTRevision = tonumber(GetAddOnMetadata("BlackMarketTracker", "Version"));
BMTRevision = "45"
if string.len(BMTRevision) == 18 then
	BMTRevision = 0;
else
	BMTRevision = tonumber(BMTRevision)
end
-- current data structure version
local BMTDataStructVer = 1
-- addon name
local BMT_ADDON_NAME = "Black Market Tracker";
-- addon name without spaces
local BMT_ADDON_NAME_NO_SPACES = "BlackMarketTracker";
-- title text on top of the browser window
local BMT_TITLE = "Black Market Tracker";
-- new addon version text shown at the bottom of the browser
local BMT_NEW_VERSION = "New addon version available";
-- coming soon text
local BMT_COMING_SOON = "Coming Soon!";
-- more than 1 week ago text (last update)
local BMT_MORE_THAN_WEEK = "More than 1 week";
-- less than 1 minute ago text (last update)
local BMT_LESS_THAN_MINUTE = "Less than 1 minute";
-- completed auction text
local BMT_COMPLETED = "Completed!";
-- alpa for fading bottom text, goes above 1.0 so that the text stays lit for a while
local BMT_TEXT_ALPHA = 5.0;
-- time of fade in/out
local BMT_TEXT_TIME = 2;
-- time for syncing the update times upon logon, in seconds
local BMT_SYNC_TIME_DELAY = 7;
-- time for syncing the items, in seconds
local BMT_SYNC_ITEMS_DELAY = 3;
-- time between window refreshes
local BMT_WINDOW_REFRESH_DELAY = 5;
-- maximum length of addon message. This does not include prefix or the char that delimits prefix and message - WoWWiki is wrong.
local BMT_MAX_ADDON_MSG_LEN = 255;
-- time after which auction will be removed from current db if no updates were done during that time and we got another item with same marketID
-- 48h
local BMT_ITEM_REMOVE_TIME_THRESHOLD = 2880;
-- predicted time left constants (time in minutes) - completed, short, medium, long, very long
local BMT_TIME_LEFT = {0, 30, 120, 720, 2880};
-- times left
local TIME_LEFT_COMPLETED = 0
local TIME_LEFT_LESS_THAN_30M = 1
local TIME_LEFT_30M_2H = 2
local TIME_LEFT_2H_12H = 3
local TIME_LEFT_MORE_THAN_12H = 4
-- array storing current db fields that are synced
local BMTItemSyncFields = 	{ "name",	"texture",	"quantity",	"itemType",	"level",	"levelType",	"sellerName",	"minBid",	"minIncrement",	"currBid",	"numBids",	"timeLeft",	"link",	"predictedLow",	"predictedHigh"	};
local BMTIsSyncFieldNumber =	{ false,	false,		true,		false,		true,		false,			false, 			true,		true,			true,		true,		true,		false,	true,			true			};
-- fading state (bottom text)
BMTFadeState = 0;
-- selected realm to display the items in the browser (same for both current and history)
BMTSelectedRealm = "";
-- used so that OnLogin gets called only upon logging in or reloading UI
BMTAddonLoaded = false;

function BMT_Debug(msg)
	if(BMTDebug) then
		local text = "|cff00ff96"..BMT_ADDON_NAME_NO_SPACES.." (Debug): |r"..msg;
		DEFAULT_CHAT_FRAME:AddMessage(text);
	end
end

function BMT_SlashDebugHandler()
	for i=1, #BlackMarketTrackerCurrentDB do
		local name = BlackMarketTrackerCurrentDB[i].name;
		local marketID = BlackMarketTrackerCurrentDB[i].marketID;
		BMT_Debug("name="..name.." marketID="..marketID);
	end
	BMT_Debug("BMTRevision="..BMTRevision)
end

BMTSchedulerFrame:SetScript("OnUpdate", function(self, elapsedVal)
	-- throttle BMT_OnUpdate calls
	BMTElapsed = BMTElapsed + elapsedVal;
	if(BMTElapsed >= BMT_ONUPDATE_SEC) then
		BMTElapsed = 0;
		BMT_OnUpdate(self, elapsedVal);
	end
end)

function BMT_InitializeAllTables()
	local realmName = GetRealmName()
	-- initialize history db for this realm
	if BlackMarketTrackerCurrentDB == nil then
		BlackMarketTrackerCurrentDB = {}
	end
	if BlackMarketTrackerCurrentDB[realmName] == nil then
		BlackMarketTrackerCurrentDB[realmName] = {}
	end
	if BlackMarketTrackerCurrentDB[realmName]["updateInfo"] == nil then
		BlackMarketTrackerCurrentDB[realmName]["updateInfo"] = {}
	end
	if BlackMarketTrackerCurrentDB[realmName]["items"] == nil then
		BlackMarketTrackerCurrentDB[realmName]["items"] = {}
	end
	-- initialize current db for this realm
	if BlackMarketTrackerHistoryDB == nil then
		BlackMarketTrackerHistoryDB = {}
	end
	--[[ item struct here
	if BlackMarketTrackerHistoryDB[realmName] == nil then
		BlackMarketTrackerHistoryDB[realmName] = {}
	end
	]]
	if BlackMarketTrackerConfig == nil then
		BlackMarketTrackerConfig = {}
	end
	if BlackMarketTrackerMisc == nil then
		BlackMarketTrackerMisc = {}
	end
	BlackMarketTrackerMisc["DataStructVer"] = BMTDataStructVer
end

function BMT_DeleteAllData()
	BlackMarketTrackerCurrentDB = nil;
	BlackMarketTrackerHistoryDB = nil;
	BlackMarketTrackerConfig = nil;
	BlackMarketTrackerMisc = nil;
	BMT_InitializeAllTables();
	ReloadUI();
end

function BMT_SlashHandler(msg, editbox)
	local msgLen = #msg;
	-- TODO: Config
	-- if any popup window is visible, ignore everything
	if BMTIsAnyPopupWindowShown() then return end
	
	if(msgLen == 0 or msg == "show" or msg == "ui" or msg == "open") then
		BMTFrame_Toggle();
	elseif(msg == "debug") then
		BMT_SlashDebugHandler();
	elseif(msg == "delete" or msg=="purge" or msg=="clear") then
		BMTShowPopupWindowConfirm(BMT_CONFIRM_WINDOW_STATE_DELETE_ALL, "Are you sure you want to delete all wipe all of your data and settings?\nNOTE: Pressing 'Yes' will also reload your UI.")
	end
end

function BMT_EnableEscInteractivity(frameName)
	local found = false
	for i=1, #UISpecialFrames do
		if UISpecialFrames[i] == frameName then
			found = true
			break
		end
	end
	if not found then 
		table.insert(UISpecialFrames, frameName)
	end
end

function BMT_DisableEscInteractivity(frameName)
	for i=1, #UISpecialFrames do
		if UISpecialFrames[i] == frameName then
			table.remove(UISpecialFrames, i)
			break
		end
	end
end

function BMTFrame_OnLoad(self)
	SLASH_BLACKMARKETTRACKER1 = "/bmt";
	SLASH_BLACKMARKETTRACKER2 = "/blackmarkettracker";
	SlashCmdList["BLACKMARKETTRACKER"] = BMT_SlashHandler;
	self:RegisterForDrag("LeftButton");
	BMT_EnableEscInteractivity(self:GetName())
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("CHAT_MSG_ADDON");
	self:RegisterEvent("BLACK_MARKET_ITEM_UPDATE");
	BMTScrollFrame.update = BMTScrollFrame_Update;
	BMTScrollFrame.scrollBar.doNotHide = true;
	BMTFrameTitleText:SetText(BMT_TITLE.." (r"..BMTRevision..")");
	RegisterAddonMessagePrefix("BMT");
end

function BMTFrame_OnEvent(self, event, ...)
	if(event == "BLACK_MARKET_ITEM_UPDATE") then
		BMTTaskManager_Unschedule(BMT_NewSearchHandler);
		BMTTaskManager_Schedule(0.5, BMT_NewSearchHandler);
	elseif(event == "CHAT_MSG_ADDON") then
		local prefix = select(1, ...);
		local msg = select(2, ...);
		local channel = select(3, ...);
		local sender = select(4, ...);
		BMTSyncing_AddonMsgReceiveHandler(prefix, msg, channel, sender);
	elseif(event == "PLAYER_ENTERING_WORLD") then
		BMTTaskManager_Unschedule(BMT_OnLoginHandler);
		BMTTaskManager_Schedule(2, BMT_OnLoginHandler);
	end
end

function BMTFrame_Toggle()
	if(BMTFrame:IsShown()) then
		BMTFrame:Hide();
		BMTTaskManager_Unschedule(BMT_FullWindowRefresh);
		--BlackMarketTrackerBottom:SetAlpha(0);
		-- TODO: sound enable toggle
	else
		BMTFrame:Show();
		-- TODO: sound enable toggle
	end
end

-- under development overlay and buttons disable
function BMTFrame_MakeUnusable()
	BMTFrame.UnderDevelopment:Show()
	BMTFrame.UnderDevelopment:EnableMouse(true)
	BMTFrame:EnableMouse(false);
	local buttons = BMTScrollFrame.buttons;
	local numButtons = 0;
	if buttons ~= nil then
		numButtons = #buttons;
	end
	for i=1, numButtons do
		buttons[i]:EnableMouse(false);
		buttons[i].Item:EnableMouse(false);
		buttons[i].TimeLeft:EnableMouse(false);
	end
end

function BMTFrame_OnShow()
	if(BMTHighestRevision <= BMTRevision) then
		BMTFadeState = 2;
	else
		BMTFadeState = 0;
	end
	BMTBottomText_FadingHandler()
	PlaySound("AuctionWindowOpen");
	HybridScrollFrame_CreateButtons(BMTScrollFrame, "BMTItemTemplate", 5, -5);
	-- under development overlay
	if not BMTDebug then
		BMTFrame_MakeUnusable()
	end
	BMT_RefreshAllPredictedTimes(BMTSelectedRealm)
	BMTScrollFrame_Update();
	-- if careful about changing this number
	if BMTSyncingState == 0 then
		BMTFrameSpinningInset:Hide()
	else
		BMTFrameSpinningInset:Show()
	end
end

function BMTFrame_OnHide()
	PlaySound("AuctionWindowClose");
	BMTTaskManager_Unschedule(BMT_FullWindowRefresh);
end

-- delay_time, function, function_parameters
function BMTTaskManager_Schedule(delay, func, ...)
	-- add params
	local task = {...};
	-- add function and time when it should be called
	task.func = func;
	task.time = GetTime() + delay;
	table.insert(BMTScheduledTasks, task);
end

function BMTTaskManager_Unschedule(func, ...)
	local entry, found;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.func == func) then
			found = true;
			for j=1, select("#", ...) do
				if(select(j, ...) ~= entry[i]) then
					found = false;
					break;
				end
			end
			if(found) then
				table.remove(BMTScheduledTasks, i);
			end
		end
	end
end

function BMT_OnUpdate(self, elapsedVal)
	-- for some reason this needs to be done
	if(BMTElapsed == 0) then return end
	
	--BMT_Debug(GetTime().." BMTElapsed="..BMTElapsed);
	local entry;
	for i=#BMTScheduledTasks, 1, -1 do
		entry = BMTScheduledTasks[i];
		if(entry.time <= GetTime()) then
			table.remove(BMTScheduledTasks, i);
			entry.func(unpack(entry));
		end
	end
end

function BMTItem_OnClick(self, button, down)
	-- Function for later use
end

function BMTBottomText_FadingHandler()
	-- -1 is only used if the window is currently opened and new version was found during that
	-- therefore we want to fade out current text first and then fade into the new one
	if(BMTFadeState == -1) then
		BMTFadeState = 0;
		BMTBottomText_FadeOut();
	elseif(BMTFadeState == 0) then
		BMTFadeState = 1;
		BMTFrameBottomText:SetText(BMT_NEW_VERSION .. " (r" .. BMTHighestRevision .. ")");
		BMTTrackerBottomText_FadeIn();
	elseif(BMTFadeState == 1) then
		BMTFadeState = 2;
		BMTBottomText_FadeOut();
	elseif(BMTFadeState == 2) then
		BMTFadeState = 3;
		local bottomText = "Last updated by ";
		if(BlackMarketTrackerCurrentDB[BMTSelectedRealm] and BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]) then
			if(not BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["playerName"] or not BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["playerClass"] or not BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["playerGuild"] or not BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateTime"] or not BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateDate"]) then
				bottomText = bottomText .. "|cffffd100Unknown|r";
			else
				local classColor = RAID_CLASS_COLORS[BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["playerClass"]];
				bottomText = bottomText .. "|cff" .. string.format("%2.2x", classColor.r*255) .. string.format("%2.2x", classColor.g*255) .. string.format("%2.2x", classColor.b*255);
				bottomText = bottomText .. BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["playerName"] .. "|r (";
				bottomText = bottomText .. BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["playerGuild"] .. "), ";
				local updateTimeStr =  string.format("%2.2d:%2.2d", (BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateTime"]/60), (BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateTime"]%60));
				bottomText = bottomText .. BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateDate"] .. " " .. updateTimeStr.." (";
				local currHour, currMinute = GetGameTime();
				local currTime = 60*currHour+currMinute;
				local diff = BMT_CalculateDateAndTimeDiff(BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateTime"], date("%d/%m/%y"), currTime);
				if(diff > 10080) then
					bottomText = bottomText .. BMT_MORE_THAN_WEEK;
				elseif(diff < 1) then
					bottomText = bottomText .. BMT_LESS_THAN_MINUTE;
				else
					local addSpace = false;
					local days = math.floor(diff/1440);
					local hours = math.floor((diff-(days*1440))/60);
					local minutes = math.floor(diff%60);
					if(days ~= 0) then
						addSpace = true;
						bottomText = bottomText .. tostring(days);
						if(days == 1) then
							bottomText = bottomText .. " day";
						else
							bottomText = bottomText .. " days";
						end						
					end
					
					if(hours ~= 0) then
						if(addSpace) then
							bottomText = bottomText .." ";
						end
						addSpace = true;
						bottomText = bottomText .. tostring(hours);
						if(hours == 1) then
							bottomText = bottomText .. " hour";
						else
							bottomText = bottomText .. " hours";
						end						
					end
					
					if(minutes ~= 0) then
						if(addSpace) then
							bottomText = bottomText .." ";
						end
						addSpace = true;
						bottomText = bottomText .. tostring(minutes);
						if(minutes == 1) then
							bottomText = bottomText .. " minute";
						else
							bottomText = bottomText .. " minutes";
						end						
					end
				end
				bottomText = bottomText .. " ago)";
			end
		else
			bottomText = bottomText .. "|cffffd100Unknown|r";
		end
		BMTFrameBottomText:SetText(bottomText);
		BMTTrackerBottomText_FadeIn();
	end
end

function BMTTrackerBottomText_FadeIn()
	local fadeInfo = {};
    fadeInfo.mode = "IN";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.endAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BMTFrameBottom, fadeInfo);
end

function BMTBottomText_FadeOut()
	local fadeInfo = {};
    fadeInfo.mode = "OUT";
    fadeInfo.timeToFade = BMT_TEXT_TIME;
	fadeInfo.startAlpha = BMT_TEXT_ALPHA;
    fadeInfo.finishedFunc = BMTBottomText_FadingHandler;
    UIFrameFade(BMTFrameBottom, fadeInfo);
end

function BMTScrollFrame_Update()
	local numItems = BMT_GetNumItems(BMTSelectedRealm)
	local scrollFrame = BMTScrollFrame;
	local offset = HybridScrollFrame_GetOffset(scrollFrame);
	local buttons = scrollFrame.buttons;
	local numButtons = #buttons;
	
	for i=1, numButtons do
		local button = buttons[i];
		local index = offset + i;
		if ( index <= numItems ) then
			local name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh = BMT_GetFromCurrentDB(index);
			if(name) then
				button.Name:SetText(name);
                button.Item.IconTexture:SetTexture(texture);
				-- config?
				--[[
                if ( not usable ) then
                    button.Item.IconTexture:SetVertexColor(1.0, 0.1, 0.1);
                else
                    button.Item.IconTexture:SetVertexColor(1.0, 1.0, 1.0);
                end
				]]
                button.Item.Count:SetText(quantity);
                button.Item.Count:SetShown(quantity > 1);
                button.Type:SetText(itemType);
				button.NumBids:SetText(numBids);
                button.Seller:SetText(sellerName);
                button.Level:SetText(level);
                local bidAmount = currBid;
                local minNextBid = currBid + minIncrement;
                if ( currBid == 0 ) then
                    bidAmount = minBid;
                    minNextBid = minBid;
                end
                MoneyFrame_Update(button.CurrentBid, bidAmount);
                button.minNextBid = minNextBid;
                button.auctionCompleate = (timeLeft == 0);
                button.TimeLeft.Text:SetText(_G["AUCTION_TIME_LEFT"..timeLeft]);
                button.TimeLeft.tooltip = _G["AUCTION_TIME_LEFT"..timeLeft.."_DETAIL"];
                button.itemLink = link;
				local predictedText = BMT_PredictedTimeGetText(predictedLow, predictedHigh);
				button.PredictedTimeLeft:SetText(predictedText);
				button.Warning:SetText(BMT_COMING_SOON);
                button:Show();
            else
                button:Hide();
			end
		else
			button:Hide();
		end
	end
	local totalHeight = numItems * scrollFrame.buttonHeight;
    local displayedHeight = numButtons * scrollFrame.buttonHeight;
    HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

function BMT_PredictedTimeGetText(predictedLow, predictedHigh)
	local text = "";
	local predictedLowH = math.floor(predictedLow/60);
	local predictedLowM = predictedLow%60;
	local predictedHighH = math.floor(predictedHigh/60);
	local predictedHighM = predictedHigh%60;
	-- more than x
	if(predictedHigh > (BMT_TIME_LEFT[#BMT_TIME_LEFT-1]+1)) then
		text = "> ";
		if(predictedLowH ~= 0) then
			text = text..tostring(predictedLowH).."h";
		end
		if(predictedLowM ~= 0) then
			if(predictedLowH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedLowM).."m";
		end
	-- less than x
	elseif(predictedLow <= 0 and predictedHigh > 0) then
		text = "< ";
		if(predictedHighH ~= 0) then
			text = text..tostring(predictedHighH).."h";
		end
		if(predictedHighM ~= 0) then
			if(predictedHighH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedHighM).."m";
		end
	-- completed
	elseif(predictedLow <= 0 and predictedHigh <= 0) then
		text = BMT_COMPLETED;
	-- between x and y
	else
		if(predictedLowH ~= 0) then
			text = text..tostring(predictedLowH).."h";
		end
		if(predictedLowM ~= 0) then
			if(predictedLowH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedLowM).."m";
		end
		text = text .. " - ";
		if(predictedHighH ~= 0) then
			text = text..tostring(predictedHighH).."h";
		end
		if(predictedHighM ~= 0) then
			if(predictedHighH ~= 0) then text = text .. " "; end
			text = text..tostring(predictedHighM).."m";
		end
	end
	return text;
end

function BMT_SetNewUpdater(realmName, playerName, playerClass, playerGuild, updateTime, updateDate)
	-- checks if table exists should be done before already
	BlackMarketTrackerCurrentDB[realmName]["updateInfo"]["playerName"] = playerName
	BlackMarketTrackerCurrentDB[realmName]["updateInfo"]["playerClass"] = playerClass
	BlackMarketTrackerCurrentDB[realmName]["updateInfo"]["playerGuild"] = playerGuild
	BlackMarketTrackerCurrentDB[realmName]["updateInfo"]["updateTime"] = updateTime
	BlackMarketTrackerCurrentDB[realmName]["updateInfo"]["updateDate"] = updateDate
	-- if tracker is opened already, update the bottom text
	if(BMTFrame:IsShown()) then
		BMTFadeState = 2
		BMTBottomText_FadeOut()
	end
end

function BMT_GetFromCurrentDB(index)
	if BlackMarketTrackerCurrentDB ~= nil and BlackMarketTrackerCurrentDB[BMTSelectedRealm] ~= nil and BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"] ~= nil and BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index] ~= nil then
		return BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].name, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].texture, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].quantity, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].itemType, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].level, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].levelType, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].sellerName, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].minBid, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].minIncrement, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].currBid, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].youHaveHighBid, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].numBids, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].timeLeft, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].link, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].marketID, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].predictedLow, BlackMarketTrackerCurrentDB[BMTSelectedRealm]["items"][index].predictedHigh;
	else return nil
	end
end

-- verifies if given number of fields can be sent through the SendAddonMessage() with given db index and starting field number. Returns the end field number.
-- item sync messages in format: BMT_SYNC_ITEM_KEY%marketID%startField%endField%Field(startField)%Field(startField+1)%...%Field(endField)
-- where % is the separator delimiting the fields
function BMT_CalculateSyncFields(index, startField)
	local numFields = #BMTItemSyncFields;
	local length = string.len(BMT_SYNC_ITEM_KEY) + string.len(tostring((BlackMarketTrackerCurrentDB[index].marketID))) + string.len(tostring(startField)) + 3*string.len(BMT_SEPARATOR);
	local endField;
	for endField=startField, numFields do
		-- add end field number string
		length = length + string.len(tostring(endField));
		-- add separator
		length = length + string.len(BMT_SEPARATOR);
		-- add field contents
		length = length + string.len(tostring(BlackMarketTrackerCurrentDB[index][BMTItemSyncFields[endField]]));
		-- string too long
		if(length > BMT_MAX_ADDON_MSG_LEN) then
			if(endField == startField) then
				-- not even one field can be sent
				return -1;
			else
				return (endField-1);
			end
		end
		-- do this before the next iteration
		length = length - string.len(tostring(endField));
	end
	
	return numFields;
end

-- TODO: Friends
function BMT_SendCurrentDB()
	if(not BlackMarketTrackerCurrentDB or not BlackMarketTrackerMisc) then return end
	if(not BlackMarketTrackerMisc.playerName) then return end
	
	local currDBSize = #BlackMarketTrackerCurrentDB;
	local numFields = #BMTItemSyncFields;
	for i=1, currDBSize do
		local name, texture, quantity, itemType, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID, predictedLow, predictedHigh = BMT_GetFromCurrentDB(i);
		local startField = 1;
		while(startField <= numFields) do
			local endField = BMT_CalculateSyncFields(i, startField);
			if(endField == -1) then 
				BMT_Debug("Unable to send field="..startField.." index="..index.." value="..BlackMarketTrackerCurrentDB[index][BMTItemSyncFields[startField]]);
				break;
			end
			local msg = BMT_SYNC_ITEM_KEY..BMT_SEPARATOR..marketID..BMT_SEPARATOR..tostring(startField)..BMT_SEPARATOR..tostring(endField);
			for j=startField, endField do
				msg = msg..BMT_SEPARATOR..BlackMarketTrackerCurrentDB[i][BMTItemSyncFields[j]];
			end
			startField = endField+1;
			--SendAddonMessage("BMT", msg, "GUILD");
		end
	end
	--SendAddonMessage("BMT", BMT_SYNC_UPDATER_KEY..BMT_SEPARATOR..BlackMarketTrackerMisc.playerName..BMT_SEPARATOR..BlackMarketTrackerMisc.playerClass..BMT_SEPARATOR..BlackMarketTrackerMisc.playerGuild..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateTime..BMT_SEPARATOR..BlackMarketTrackerMisc.UpdateDate, "GUILD");
	--SendAddonMessage("BMT", BMT_SYNC_ITEMEND_KEY, "GUILD");
end

-- recalculates predicted times left for item under [index] for [realmName] realm
-- this also sets the calculated predicted times, predictionsUpdateDate and predictionsUpdateTime for that item
-- DOES NOT set other fields like timeLeft to new value
-- --parameters:
-- --realmName - name of the realm of a given item
-- --index - index of the item in realm's item db
-- --newPredictedLow, newPredictedHigh, newTimeLeft, newUpdateDate, newUpdateTime - correspond to the new item
-- --fromWindwoRefresh - true if this function has been called by just refreshing the browser window, false otherwise
-- this can be called either to recalculate the times upon browsing BMAH ourself, receiving a sync or just refreshing browser list (in which case the same timeLeft should be passed and fromWindowRefresh set to true)
-- upon receiving a sync 'negotiating' sync (finding narrowest possible prediction) this function will recalculate predictions, as if receiver was the one searching the items
function BMT_RecalculateAndSetPredictedTimes(realmName, index, newTimeLeft, fromWindowRefresh)
	local prevPredictedLow = BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedLow"]
	local prevPredictedHigh = BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedHigh"]
	local prevTimeLeft = BlackMarketTrackerCurrentDB[realmName]["items"][index]["timeLeft"]
	local prevUpdateDate = BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictionsUpdateDate"]
	local prevUpdateTime = BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictionsUpdateTime"]
	local currDate = date("%d/%m/%y")
	local currHour, currMinute = GetGameTime()
	local currTime = 60*currHour+currMinute
	local updatedMinAgo = BMT_CalculateDateAndTimeDiff(currDate, currTime, prevUpdateDate, prevUpdateTime)
	-- set those fields here already, as this should be done for all cases
	BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictionsUpdateDate"] = currDate
	BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictionsUpdateTime"] = currTime
	-- completed case
	if newTimeLeft == TIME_LEFT_COMPLETED then
		BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedLow"] = BMT_TIME_LEFT[1] -- 0
		BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedHigh"] = BMT_TIME_LEFT[1] -- 0
		return
	end
	-- more than 12h cas
	if newTimeLeft == TIME_LEFT_MORE_THAN_12H then return end
	-- more than completed, less than 12h
	-- if timeLeft changed, example:
	-- old: 2h-12h
	-- new: 30m-2h
	-- ..............30m.....................2h............................................12h	scale
	-- ......................................|<--------------old prediction------------>|....	old prediction
	-- .........................|<--------------old prediction------------>|.................	old prediction moved left by (updatedMinAgo=newUpdateTime-oldUpdateTime)
	-- .........................|<----new--->|...............................................	everything going under lower time range (30m) or above higher time range (2h) is cut off
	
	if newTimeLeft < prevTimeLeft then
		BMT_Debug("index="..index.." name="..BlackMarketTrackerCurrentDB[realmName]["items"][index]["name"].." UP->DOWN")
		-- never touch the upper range of 'more than 12h'
		if prevTimeLeft ~= TIME_LEFT_MORE_THAN_12H then
			prevPredictedHigh = prevPredictedHigh - updatedMinAgo
		end
		-- lower range can go to negative, it will get cut re-adjusted anyway
		prevPredictedLow = prevPredictedLow - updatedMinAgo
		-- cut everything that goes under or above given (blizzard's) time left range
		if prevPredictedLow < BMT_TIME_LEFT[newTimeLeft] then
			prevPredictedLow = BMT_TIME_LEFT[newTimeLeft]
		end
		if prevPredictedHigh > BMT_TIME_LEFT[newTimeLeft+1]+1 then
			prevPredictedHigh = BMT_TIME_LEFT[newTimeLeft+1]+1
		end
		-- set the fields
		BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedLow"] = prevPredictedLow
		BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedHigh"] = prevPredictedHigh
		return
	end
	
	-- if timeLeft didn't change or we're jst refreshing the window, then we just cut (updatedMinAgo=newUpdateTime-oldUpdateTime) from the upper range
	if newTimeLeft == prevTimeLeft then
		-- never touch the upper range of 'more than 12h'
		if prevTimeLeft ~= TIME_LEFT_MORE_THAN_12H then
			prevPredictedHigh = prevPredictedHigh - updatedMinAgo
		end
		-- if we're just opening the BMT window, then we lower both, if searching we cut the lower range to lower timeLeft range
		prevPredictedLow = prevPredictedLow - updatedMinAgo
		if not fromWindowRefresh then
			if prevPredictedLow < BMT_TIME_LEFT[newTimeLeft] then
				prevPredictedLow = BMT_TIME_LEFT[newTimeLeft]
			end
		end
		BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedLow"] = prevPredictedLow
		BlackMarketTrackerCurrentDB[realmName]["items"][index]["predictedHigh"] = prevPredictedHigh
		-- prevPredictedLow stays on the lower range of current given (blizzard's) time left
		return
	end
end

-- called only when refreshing browser, otherwise BMT_RecalculateAndSetPredictedTimes should be called directly for each item
function BMT_RefreshAllPredictedTimes(realmName)
	local numItems = BMT_GetNumItems(BMTSelectedRealm)
	for i=1, numItems do
		local currTimeLeft = BlackMarketTrackerCurrentDB[realmName]["items"][i]["timeLeft"]
		local currDate = date("%d/%m/%y");
		local currHour, currMinute = GetGameTime();
		local currTime = 60*currHour+currMinute;
		BMT_RecalculateAndSetPredictedTimes(realmName, i, currTimeLeft, true)
	end
end

-- moves all outdated entries to history DB and removes them from current DB under given criteria:
-- a - date and time of last timeLeft change is greater than timeLeftHighRange+1h (for timeLeft >12h assuming that auction can't ever be up for longer than 49h)
-- b - auction with given marketID cannot be found among auctions from searching/syncing
-- c - auction with given marketID can be found among auctions from searching/syncing but has a different name or its time left went up (assuming auctions' time left don't get extended upon bidding) or gold went down or number of bids went down
----- parameters are:	- realmName to remove the items for
--						- foundItems - list of all items found by searching BMAH or from syncing
-- performs removal from current table
-- initializes current and history tables for a given realm if needed
function BMT_MoveAllOldEntriesToHistory(realmName, foundItems)
	-- initialize it here anyway, as this always done before adding new items, therefore we won't have to do it later
	if BlackMarketTrackerCurrentDB[realmName] == nil then
		BlackMarketTrackerCurrentDB[realmName] = {}
	end
	if BlackMarketTrackerCurrentDB[realmName]["items"] == nil then
		BlackMarketTrackerCurrentDB[realmName]["items"] = {}
	end
	if BlackMarketTrackerCurrentDB[realmName]["updateInfo"] == nil then
		BlackMarketTrackerCurrentDB[realmName]["updateInfo"] = {}
	end
	-- TODO: initialize history DB
	
	-- criterion a
	
	-- criteria b and c
	local currDBSize = #BlackMarketTrackerCurrentDB[realmName]["items"]
	local i=1
	while i<=currDBSize do
		local found = false
		local nameMatches = true
		local timeLeftWentUp = false
		local goldWentDown = false
		local numBidsWentDown = false
		for j=1, #foundItems do
			if BlackMarketTrackerCurrentDB[realmName]["items"][i]["marketID"] == foundItems[j]["marketID"] then
				found = true
				if BlackMarketTrackerCurrentDB[realmName]["items"][i]["name"] ~= foundItems[j]["name"] then
					nameMatches = false
				end
				if BlackMarketTrackerCurrentDB[realmName]["items"][i]["timeLeft"] < foundItems[j]["timeLeft"] then
					timeLeftWentUp = true
				end
				if BlackMarketTrackerCurrentDB[realmName]["items"][i]["currBid"] > foundItems[j]["currBid"] then
					goldWentDown = true
				end
				if BlackMarketTrackerCurrentDB[realmName]["items"][i]["numBids"] > foundItems[j]["numBids"] then
					numBidsWentDown = true
				end
				break
			end
		end
		--BMT_Debug("i="..i.." found="..tostring(found).." nameMatches="..tostring(nameMatches).." timeLeftWentUp="..tostring(timeLeftWentUp).." goldWentDown="..tostring(goldWentDown))
		if not found then
			BMT_MoveEntryToHistory(realmName, i)
			currDBSize=currDBSize-1
			i=i-1
		end
		if found and (not nameMatches or timeLeftWentUp or goldWentDown or numBidsWentDown) then
			BMT_MoveEntryToHistory(realmName, i)
			currDBSize=currDBSize-1
			i=i-1
		end
		i=i+1
	end
end

function BMT_MoveEntryToHistory(realmName, index)
	BMT_Debug("TODO: Moving  "..BlackMarketTrackerCurrentDB[realmName]["items"][index]["name"].." for realm "..realmName.." to history DB")
	table.remove(BlackMarketTrackerCurrentDB[realmName]["items"], index)
end

-- called only when receiving a force update sync or searching for items outself (after which we might get a force update sync that will narrow predicted times even further)
function BMT_AddOrUpdateExisting(realmName, foundItems, fromSync)
	for i=1, #foundItems do
		local found = false
		for j=1, #BlackMarketTrackerCurrentDB[realmName]["items"] do
			-- no need to check if names etc. are the same, as it has been done before
			-- therefore, if we've found two items with the same marketID, we are sure that this is still the same item
			-- if something breaks here, check BMT_MoveAllOldEntriesToHistory criterias
			if foundItems[i]["marketID"] == BlackMarketTrackerCurrentDB[realmName]["items"][j]["marketID"] then
				found = true
				-- if receiving a sync, we trust it (skip the ifs)
				-- if searching ourself, we need to set the highest bidder again if gold didn't change (otherwise it will reset to "Unknown")
				-- if we are the highest bidder, none of those cases will be true, therefore foundItems[i]["highestBidder"] will
				-- already have our name
				if not fromSync then
					if foundItems[i]["currBid"] == BlackMarketTrackerCurrentDB[realmName]["items"][j]["currBid"] then
						foundItems[i]["highestBidder"] = BlackMarketTrackerCurrentDB[realmName]["items"][j]["highestBidder"]
					end
				end
				-- other fields that we don't want overriden by new/default should go here
				-- if timeLeft changed, change those fields too
				if foundItems[i]["timeLeft"] < BlackMarketTrackerCurrentDB[realmName]["items"][j]["timeLeft"] then
					foundItems[i]["lastTimeLeftChangeDate"] = currDate
					foundItems[i]["lastTimeLeftChangeTime"] = currTime
				-- otherwise keep their old values
				else
					foundItems[i]["lastTimeLeftChangeDate"] = BlackMarketTrackerCurrentDB[realmName]["items"][j]["lastTimeLeftChangeDate"]
					foundItems[i]["lastTimeLeftChangeTime"] = BlackMarketTrackerCurrentDB[realmName]["items"][j]["lastTimeLeftChangeTime"]
				end
				-- recalculate predicted times here
				BMT_RecalculateAndSetPredictedTimes(realmName, j, foundItems[i]["timeLeft"], false)
				-- rewrite predicted times to foundItems table as it still has the default ones
				foundItems[i]["predictedLow"] = BlackMarketTrackerCurrentDB[realmName]["items"][j]["predictedLow"]
				foundItems[i]["predictedHigh"] = BlackMarketTrackerCurrentDB[realmName]["items"][j]["predictedHigh"]
				-- finally set all the fields in the DB
				BlackMarketTrackerCurrentDB[realmName]["items"][j] = foundItems[i]
			end
		end
		-- adding new
		if found == false then
			BMT_Debug("Adding new entry: "..foundItems[i]["name"])
			BlackMarketTrackerCurrentDB[realmName]["items"][#BlackMarketTrackerCurrentDB[realmName]["items"]+1] = foundItems[i]
		end
	end
end

-- called only when searching BMAH by yourself
function BMT_NewSearchHandler()
	-- check if Blizzard's black market ui is still opened
	if(not BlackMarketFrame:IsShown()) then return end
	local realmName = GetRealmName();
	local playerName = UnitName("player");
	local _, playerClass = UnitClass("player");
	local playerGuild = GetGuildInfo("player");
	local currHour, currMinute = GetGameTime();
	local currTime = 60*currHour+currMinute;
	local currDate = date("%d/%m/%y");
	local numItems = C_BlackMarket.GetNumItems();
	if(numItems == nil) then return end
	
	local updatedMinAgo = 0;
	-- throttling here, ignoring searches done less than 1m ago
	if BlackMarketTrackerCurrentDB[BMTSelectedRealm] and BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"] then
		if BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateDate"] and BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateTime"] then
			updatedMinAgo = BMT_CalculateDateAndTimeDiff(BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateDate"], BlackMarketTrackerCurrentDB[BMTSelectedRealm]["updateInfo"]["updateTime"], date("%d/%m/%y"), currTime)
		end
	end
	
	
	-- no need to update this realm's table structure
	local foundItems = {}
	for i=1, numItems do
		local name, texture, quantity, itemType, _, level, levelType, sellerName, minBid, minIncrement, currBid, youHaveHighBid, numBids, timeLeft, link, marketID = C_BlackMarket.GetItemInfoByIndex(i)
		foundItems[i] = {}
		foundItems[i]["name"] = name
		foundItems[i]["texture"] = texture
		foundItems[i]["quantity"] = quantity
		foundItems[i]["itemType"] = itemType
		foundItems[i]["level"] = level
		foundItems[i]["levelType"] = levelType
		foundItems[i]["sellerName"] = sellerName
		foundItems[i]["minBid"] = minBid
		foundItems[i]["minIncrement"] = minIncrement
		foundItems[i]["currBid"] = currBid
		local highestBidder = "Unknown"
		if youHaveHighBid then
			highestBidder = UnitName("player")
		end
		foundItems[i]["highestBidder"] = highestBidder
		foundItems[i]["numBids"] = numBids
		foundItems[i]["timeLeft"] = timeLeft
		foundItems[i]["link"] = link
		foundItems[i]["marketID"] = marketID
		foundItems[i]["predictedLow"], foundItems[i]["predictedHigh"] = BMT_PredictedTimeGetDefaults(timeLeft)
		-- update date and time of predicted times
		foundItems[i]["predictionsUpdateDate"] = currDate
		foundItems[i]["predictionsUpdateTime"] = currTime
		-- date and time of last blizzard's timeLeft change (used to find outdated auctions)
		foundItems[i]["lastTimeLeftChangeDate"] = currDate
		foundItems[i]["lastTimeLeftChangeTime"] = currTime
	end
	
	-- if we have same amount of items and last update was done less than 1m ago then throttle
	-- there could be more conditions checked, like if any timeLeft/numBids/currBid changed, but for now this will do
	-- NOTE: numItems here is blizzard's numItems, our is returned by BMT_GetNumItems()
	if updatedMinAgo < 1 and numItems == BMT_GetNumItems(realmName) then return end
	
	-- move old entries to history
	BMT_MoveAllOldEntriesToHistory(realmName, foundItems)
	
	-- at this point we should have no old items in current db, update existing or add new ones
	BMT_AddOrUpdateExisting(realmName, foundItems, false)
	
	BMT_SetNewUpdater(realmName, playerName, playerClass, playerGuild, currTime, currDate);
	-- update the list in case the addon's browser is opened
	-- bottom text is already updated in BMT_SetNewUpdater()
	if(BMTFrame:IsShown()) then
		BMTScrollFrame_Update();
	end
	-- sync to others
	--BMT_SendCurrentDB();
end

function BMT_GetNumItems(realmName)
	if BlackMarketTrackerCurrentDB ~= nil and BlackMarketTrackerCurrentDB[realmName] ~= nil and BlackMarketTrackerCurrentDB[realmName]["items"] ~= nil then
		return #BlackMarketTrackerCurrentDB[realmName]["items"]
	else return 0
	end
end

function BMT_PredictedTimeGetDefaults(timeLeft)
	-- completed
	if(timeLeft == TIME_LEFT_COMPLETED) then
		return 0,0;
	else
		-- example: timeLeft=1 is 0-30m, BMT_TIME_LEFT[1]=0, BMT_TIME_LEFT[2]=30
		return BMT_TIME_LEFT[timeLeft], BMT_TIME_LEFT[timeLeft+1]+1;
	end
end

-- returns difference between given poits in time in minutes (absolute)
-- date is a string in "dd/mm/yy" format, time is a number in 60*hour+minutes format
function BMT_CalculateDateAndTimeDiff(firstDate, firstTime, secondDate, secondTime)
	local t = {};
	local day, month, year = strsplit("/", firstDate);
	t["year"] = tonumber(year) + 2000;
	t["month"] = tonumber(month);
	t["day"] = tonumber(day);
	t["hour"] = firstTime / 60;
	t["min"] = firstTime % 60;
	t["sec"] = 0;
	local first = time(t);
	day, month, year = strsplit("/", secondDate);
	t["year"] = tonumber(year) + 2000;
	t["month"] = tonumber(month);
	t["day"] = tonumber(day);
	t["hour"] = secondTime / 60;
	t["min"] = secondTime % 60;
	t["sec"] = 0;
	local second = time(t);
	return (math.floor(math.abs(first-second)/60));
end

-- returns "smaller" if first date/time is smaller, "equal" if they are equal, "greater" if first is greater.
-- date is a string in "dd/mm/yy" format, time is a number in 60*hour+minutes format
function BMT_CompareDateAndTime(firstDate, firstTime, secondDate, secondTime)
	local firstDay, firstMonth, firstYear = strsplit("/", firstDate);
	local secondDay, secondMonth, secondYear = strsplit("/", secondDate);
	if(firstYear > secondYear) then return "greater"
	elseif(firstYear < secondYear) then return "smaller"
	else
		if(firstMonth > secondMonth) then return "greater"
		elseif(firstMonth < secondMonth) then return "smaller"
		else
			if(firstDay > secondDay) then return "greater"
			elseif(firstDay < secondDay) then return "smaller"
			else
				if(firstTime > secondTime) then return "greater"
				elseif(firstTime < secondTime) then return "smaller"
				else return "equal"
				end
			end
		end
	end
end

function BMT_OnLoginHandler()
	if not BMTAddonLoaded then BMTAddonLoaded = true
	else return
	end
	-- in syncing module, has to be done here, as otherwise this function cannot be seen
	BMTSpinningInset_Initialize()
	-- set selected realm to own as default
	BMTSelectedRealm = GetRealmName()
	-- check data structure version, do necessary steps if needed
	if BlackMarketTrackerMisc then
		if BlackMarketTrackerMisc.DataStructVer == nil and (BlackMarketTrackerConfig ~= nil or BlackMarketTrackerHistoryDB ~= nil or BlackMarketTrackerCurrentDB ~= nil) then
			BMTShowPopupWindowOkay(BMT_OKAY_WINDOW_STATE_OUTDATED_DATA_STRUCT, "Data structure of "..BMT_ADDON_NAME.." is outdated and needs to be re-initialized.\nNOTE: This will reload your UI and wipe your current BMT data.")
		end
	end
	BMT_InitializeAllTables()
	-- sync version
	BMTSyncing_AskForVersion_OnLogin()
end